------------------------------------------------------------------------------
-- (Static Forward) Program Slicing Tool 
-- v1.3 (Feb, 2005)
--
--
-- by German Vidal (Technical University of Valencia, Spain)
------------------------------------------------------------------------------

import Flat 
import Flat2Fcy
import List
--import FlatTools  (deprecated)
import Unsafe

------------------------------------------------------------------------------
-- some global definitions:

printVersion =
  do putStrLn "Curry Program Slicing Tool (Version 1.3 of Feb 2005)"
     putStrLn "(Technical University of Valencia)\n"

-- should some debug information be printed during partial evaluation?
pevalDebug = False
pevalDebugBrief = False

-- maximal number of variables occurring in each unfolded expression
-- (used in getMatchedRHS to produce a safe renaming):
maxExpVars = 100

-- special variable:
varBottom = Var (-1)

-- is this call a SLICE annotation?
isSliceAnnot (Comb ctype name args) =
  if ctype==FuncCall && length args == 1 && take 5 (reverse name) == "ECILS"
  then True  -- this is a call annotated with SLICE
  else False

-- get the annotated expression (provided that this is a SLICE annotation):
getPevalAnnotExpr (Comb _ _ args) = head args

------------------------------------------------------------------------------
-- data structures used in the program slicing tool:

-- data type for substitutions:
data Subst = Sub [Int] [Expr] | FSub

-- encapsulating the state during the slicing process:
-- the state contains:
-- * the list of the prelude function declarations
-- * the list of all user-defined function declarations
-- * index for renamed variables
-- * iteration number (useful for renaming of variables)

data EvalState = EState [FuncDecl] [FuncDecl] Int Int

getPreludeDecls  (EState pfds _   _ _) = pfds
getFuncDecls     (EState _    fds _ _) = fds
getRenamingIndex (EState _    _   i _) = i
getIter          (EState _    _   _ i) = i

-- update eval state with new index for renamed variables:
updateRenamingIndex :: EvalState -> Int -> EvalState
updateRenamingIndex (EState exps fds _ d) i = EState exps fds i d


------------------------------------------------------------------------------
-- Main Functions..
------------------------------------------------------------------------------

-- alternative main call: provide only the name of program
-- the slicing criterion must be marked by (SLICE ...)
-- where SLICE is the identity function (e.g., defined in the prelude)
main :: String -> IO ()
main p =
  do printVersion
     annprog <- readFlatCurry p
     let (prog,es) = getEvalExprs annprog
     main_aux_ p (removeRenamings (removeVarsLits es)) prog prog 

main_aux_ p es _ prog =
  do prelu <- readFlatCurry "xprelude"
     main_aux p es prelu prog

main_aux :: String -> [Expr] -> Prog -> Prog -> IO ()
main_aux p [] _ prog = 
  do putStrLn "There is no criterion to start program slicing..."
     putStrLn("\nWriting original program into \""++pefile++"\"...\n")
     writeFCY pefile prog
  where pefile = p++"_slice.fcy"
main_aux p (e:es) prelu prog =
  do putStrLn ("Slicing criterion:\n")
     putStrLn (concatMap (\ex->ppExpr 0 ex ++"\n") (e:es))
     putStrLn ("Starting program slicing for program \""++p++"\"...\n")
     putStrLn ("\nSlice:\n")
     --putStrLn ("\nFinal set of expressions:\n" ++ ppExprs exps ++ "\nEND")
     ppResultants resultants
     putStrLn("\nWriting slice into \""++pefile++"\"...\n")
     writeFCY pefile slice --(trace (showFlatProg slice) slice) 
  where
    exps__ = pevalProg prelu prog (e:es)
    exps_ = reverse (map fst exps__)
    exps = removeNonUserFuns prog exps_
    ----this is to avoid deleting functions 'goal' and 'bench'
    ----which are needed to take benchmark times
    ----goalbench = extractGoalBench prog
    ----exps2 = exps ++ goalbench
    --resultants_ = buildResultants prog exps
    --resultants = replaceByTop (map topFun exps) resultants_
    resultants = extractSlice prog exps exps --exps2 exps2
    slice = buildSlice prog resultants
    pefile = p++"_slice.fcy"


extractGoalBench (Prog _ _ _ funs _ _) = exGB funs

exGB [] = []
exGB (Func f arity _ _ : rest) 
  | take 4 (reverse f) == "laog" 
     = Comb FuncCall f (nvars arity 0) : exGB rest
  | take 5 (reverse f) == "hcneb" 
     = Comb FuncCall f (nvars arity 0) : exGB rest
  | otherwise = exGB rest

nvars n x = if n==0 then [] else Var x : nvars (n-1) (x+1)

--rename e = substitute vars (index2vars (map (maxExpVars+) vars)) e
--  where vars = nonLocalVarsInExp e

ppExprs [] = "\n"
ppExprs (e:es) = "\n" ++ (ppExpr 0 e) ++ (ppExprs es)

removeNonUserFuns (Prog _ _ _ funs _ _) es = remNUFuns funs es

remNUFuns _ [] = []
remNUFuns funs (e:es) =
  if isFun e then remNUFuns_ funs (e:es)
             else remNUFuns funs es

remNUFuns_ funs ((Comb FuncCall f args):es) =
  if definedFunc2 funs f
     then (Comb FuncCall f args) : (remNUFuns funs es)
     else remNUFuns funs es

buildSlice :: Prog -> [(Expr,Expr)] -> Prog
buildSlice (Prog modname imports types oldfuns ops trans) newfuns =
    Prog modname imports types finalfuns ops trans --newtrans
  where finalfuns = mergeFuns oldfuns newfuns
        --newtrans = updateTrans trans oldfuns finalfuns

mergeFuns [] _ = []
mergeFuns ( (Func f arity typef (Rule _ _)):rest) funs =
    if rhs2 == varBottom
       then mergeFuns rest funs
       else ((Func f arity typef (Rule vars2 rhs2)) : (mergeFuns rest funs))
  where (vars2,rhs2) = lookupFun f funs

lookupFun _ [] = ([],varBottom)
lookupFun f ((Comb FuncCall g args, rhs) : rest) =
  if f==g then (map (\(Var i)->i) args, rhs)
          else lookupFun f rest

---------------------------------------------------------------------------
-- extract all expressions e embedded in a call (SLICE e) in a program
-- and replace these calls by e:

getEvalExprs :: Prog -> (Prog,[Expr])
getEvalExprs (Prog modname imports types oldfuns ops trans) =
  (Prog modname imports types newfuns ops trans, evalexps)
 where
   funexppairs = map getFuncEvalExprs oldfuns
   newfuns = map fst funexppairs
   evalexps = concatMap snd funexppairs
   
getFuncEvalExprs :: FuncDecl -> (FuncDecl,[Expr])
getFuncEvalExprs (Func name arity ftype (External ename)) =
  (Func name arity ftype (External ename), [])
getFuncEvalExprs (Func name arity ftype (Rule lhs rhs)) =
  (Func name arity ftype (Rule lhs newrhs), evalexps)
 where
   (newrhs,evalexps) = getEvalExprsInExp rhs

getEvalExprsInExp (Var i) = (Var i, [])
getEvalExprsInExp (Lit l) = (Lit l, [])
getEvalExprsInExp (Comb ctype name args)
  | isSliceAnnot (Comb ctype name args)
              = let expr = getPevalAnnotExpr (Comb ctype name args)
                 in (expr,[expr])
  | otherwise = let (newargs,pexps) = getEvalExprsInExps args
                 in (Comb ctype name newargs, pexps)
getEvalExprsInExp (Apply e1 e2) =
 let (ne1,pe1) = getEvalExprsInExp e1
     (ne2,pe2) = getEvalExprsInExp e2
  in (Apply ne1 ne2, pe1++pe2)
getEvalExprsInExp (Constr vars e) =
 let (ne,pe) = getEvalExprsInExp e
  in (Constr vars ne, pe)
getEvalExprsInExp (Or e1 e2) =
 let (ne1,pe1) = getEvalExprsInExp e1
     (ne2,pe2) = getEvalExprsInExp e2
  in (Or ne1 ne2, pe1++pe2)
getEvalExprsInExp (Case ctype e bs) =
 let (ne,pe1) = getEvalExprsInExp e
     (nbs,pe2) = getEvalExprsInBranches bs
  in (Case ctype ne nbs, pe1++pe2)
getEvalExprsInExp (Choice e) =
 let (ne,pe) = getEvalExprsInExp e
  in (Choice ne, pe)
getEvalExprsInExp (GuardedExpr vars e1 e2) =
 let (ne1,pe1) = getEvalExprsInExp e1
     (ne2,pe2) = getEvalExprsInExp e2
  in (GuardedExpr vars ne1 ne2, pe1++pe2)

getEvalExprsInExps [] = ([],[])
getEvalExprsInExps (e:es) = let (ne,pe1) = getEvalExprsInExp e
                                (nes,pe2) = getEvalExprsInExps es
                             in (ne:nes, pe1++pe2)

getEvalExprsInBranches [] = ([],[])
getEvalExprsInBranches (Branch p e : bs) =
   let (ne,pe1) = getEvalExprsInExp e
       (nbs,pe2) = getEvalExprsInBranches bs
    in (Branch p ne : nbs, pe1++pe2)


---------------------------------------------------------------------------
--
-- partial evaluation of a program (only the defined functions are used)
--

pevalProg (Prog _ _ _ pfuns _ _) (Prog _ _ _ funs _ _) es = 
  pevalLoop pfuns funs 0 (buildStates es pfuns funs) []

buildStates [] _ _ = []
buildStates (e:es) pfuns funs = 
  flatState (EState pfuns funs (maxVarIndex e) 0) (e,[]) : buildStates es pfuns funs

flatState pst (e,s)
  | isBuiltin e   = (e,s) --flatStateBuiltin pst (e,s) --should we accept builtins here? only user-defined funcs?
  | isFun e       = flatStateF pst (e,s)
  | otherwise     = (e,s)

flatStateBuiltin pst (Comb FuncCall c [e1,e2],s) 
  | not (isLit e1) = flatState_ pst e1 (Comb FuncCall c [e1,e2], s)
  | not (isLit e2) = flatState_ pst e2 (Comb FuncCall c [e1,e2], s)
  | otherwise      = (Comb FuncCall c [e1,e2], s)


flatStateF pst (e,s) = 
  --if trace ("\n@@@fun_eval of:\n" ++ ppExpr 0 e ++ "\n@@@returned:\n" ++ ppExpr 0 unf ++ "\n@@@and case_eval:\n" ++ ppExprs redS ++ "\n@@@and searchIsCaseFun returns:" ++ ppExpr 0 red ++ "\n") red==varBottom
  if red==varBottom
     then (e,s)
     else flatStateC pst red (e,s)
  where unf = fun_eval pst e
        redS = findall (\r -> (case_eval pst unf =:= r))
        red = searchIsCaseFun redS

flatStateC pst (Case _ call _) state = flatState_ pst call state
--flatState_ pst (trace ("\n@@@calling flatState_ with call:\n" ++ ppExpr 0 call ++ "\n@@@and state:" ++ ppStates [state]) call) state

searchIsCaseFun [] = varBottom
searchIsCaseFun (e:es) =
  if isCaseFun e || isCaseApply e then e 
                                  else searchIsCaseFun es

flatState_ pst call (e,s) = flatState pst (call,(newe,var) : s)
  where (newe,var) = replace_expr pst call e

ppKKK (a,_) = ppExpr 0 a

replace_expr pst call e =
  let curr_ri = getRenamingIndex pst
      n = getIter pst
      freshIndex = curr_ri+(maxExpVars*n)+1
  in
    if call==e then (Var freshIndex, freshIndex)
               else replace_expr_ pst call e
               --else replace_expr_ pst call (trace ("\n@@@@@calling replace_expr_ with: " ++ ppExpr 0 e ++ "\n") e)

--
replace_expr_ _ _ (Var x) = (Var x, -1)

replace_expr_ _ _ (Lit l) = (Lit l, -1)

replace_expr_ pst call (Comb ftype f es) =
  (Comb ftype f new_es, var)
  where (new_es,var) = replace_expr_args pst call [] es

replace_expr_ pst call (Case ctype e ces) =
  if newe==e then replace_expr_branches pst call [] ces ctype e 
             else (Case ctype newe ces, var)
  where (newe,var) = replace_expr pst call e

replace_expr_ pst call (Apply e1 e2) = (Apply ne1 ne2, var)
  where ([ne1,ne2],var) = replace_expr_args pst call [] [e1,e2]

replace_expr_ pst call (Or e1 e2) = (Or ne1 ne2, var)
  where ([ne1,ne2],var) = replace_expr_args pst call [] [e1,e2]

replace_expr_ pst call (GuardedExpr vars gc e) = 
  (GuardedExpr vars ngc ne, var)
  where ([ngc,ne],var) = replace_expr_args pst call [] [gc,e]

replace_expr_ pst call (Constr vars e) = 
  (Constr vars ne, var)
  where ([ne],var) = replace_expr_args pst call [] [e]

replace_expr_ pst call (Choice e) = (Choice ne, var)
  where ([ne],var) = replace_expr_args pst call [] [e]

--
replace_expr_args _ _ old [] = (old,-1)
replace_expr_args pst call old (e:es) =
  if newe==e then replace_expr_args pst call (old++[e]) es
             else (old++(newe:es), var)
  where (newe,var) = replace_expr pst call e


replace_expr_branches _ _ old [] ctype arg = (Case ctype arg old,-1)
replace_expr_branches pst call old (Branch p e : ces) ctype arg =
  if newe==e 
     then replace_expr_branches pst call (old++[Branch p e]) ces ctype arg
     else (Case ctype arg (old++((Branch p newe):ces)), var)
  where (newe,var) = replace_expr pst call e

----
pevalLoop pfuns funs n es pairs 
  | if pevalDebug
     then trace ("\n      ***Iteration " ++ (show n) 
          ++ "***\n\nCURRENT SET: " ++ (ppStates es) 
              -- ++ "OLD PAIRS: " ++ (ppPairs pairs) 
          ++ "NEW PAIRS: " ++ (ppPairs newpairs) 
          ++ "NEW SET: " ++ (ppStates new) 
          ++ "AFTER ABSTRACTION: " ++ (ppStates new_es) ++ "\n") success
     else if pevalDebugBrief 
             then trace ("\n      ***Iteration " ++ (show n) 
                  ++ "***\n\nCURRENT SET: " ++ (ppStates es)) success
             else success
  = if equalMR (map state2expr es) (map state2expr new_es)
       then es
       else pevalLoop pfuns funs (n+1) new_es newpairs
  where newpairs = pevalList pfuns funs n es pairs
        oldStates = concat (map snd pairs)
        newStates = concat (map snd newpairs)
        new = diffPairs newStates oldStates
        new_es = abstract pfuns funs n new es
        --new_es = replace_flatten pfuns funs n new_es_  --this was a bad idea!! (destroys the
                                                         --monovariant strategy, since it may return
                                                         --exprs with the same func. symbol)


-- we only need to apply abstract over the "new" elements..
diffPairs :: [a] -> [a] -> [a]
diffPairs [] _ = []
diffPairs (e:es) olds = 
  if e `elem` olds then diffPairs es olds else e : (diffPairs es olds)

pevalList _ _ _ [] _ = []
pevalList pfuns funs n (e:es) pairs =
  if spe == []
     then pevalList_ pfuns funs n (e:es) pairs
     else (e,spe) : (pevalList pfuns funs n es pairs)
  where spe = searchPair e pairs
        

--recover the stored partial evaluation of a given term (or return [])
searchPair _ [] = []
searchPair e ((x,xpe) : es) = if e==x then xpe else searchPair e es

pevalList_ pfuns funs n ((e,s):es) pairs 
  = strict (((e,s),spe) : (pevalList pfuns funs n es (((e,s),spe) : pairs)))
  where pst = EState pfuns funs (maxVarIndex (state2expr (e,s))) n
        unf = fun_eval pst e
        spe_ = findall (\r -> (case_eval pst unf =:= r))
        spe = replace_flatten pfuns funs n (addStacks spe_ s)
        --spe = replace_flatten pfuns funs n (addStacks (trace ("\n###fun_eval of:\n" ++ ppExpr 0 e ++ "\n###returned:\n" ++ ppExpr 0 unf ++ "\n###and case_eval:\n" ++ ppExprs spe_ ++ "\n") spe_) s)


--  | if pevalDebug
--     then trace ("Iteration: " ++ (show n) ++ 
--          ppStates [(e,s)] ++ " => " ++ ppStates spe ++ "\n") success
--     else success

replace_flatten _ _ _ [] = []
replace_flatten pfuns funs n (se:ses) =
    flat : replace_flatten pfuns funs n ses
  where (e,s) = first (replace_eval se)
        flat =  flatState (EState pfuns funs (maxVarIndex (state2expr (e,s))) n) 
                          (e,s)

addStacks [] _ = []
addStacks (e:es) s = (e,s) : addStacks es s

---
ppStates [] = "\n"
ppStates ((e,s):es) = 
  "\n--->Expr: " ++ (ppExpr 0 e) ++ "\n--->Stack: " ++ printStack s ++ (ppStates es)

ppPairs [] = "\n"
ppPairs ((s,ss):sss) = "\nFROM " ++ (ppStates [s]) ++ "WE GET " ++ (ppStates ss) ++ " ** " ++ (ppPairs sss)

printStack [] = "\n"
printStack ((e,x):es) =  
  "( " ++ (ppExpr 0 e) ++ " , " ++ (ppVar x) ++ ") --- " ++ (printStack es)

ppSub FSub = "FSub" ++ "\n"
ppSub (Sub vars exps) = ppSub_ vars exps

ppSub_ [] [] = "\n"
ppSub_ (v:vs) (e:es) = ppVar v ++ " --> " ++ ppExpr 0 e ++ ppSub_ vs es

-- if we replace "strict" by "id" above, the computation is lazy
-- but needs much more time (90 times longer for KMP benchmark)!
strict x | x==x = x

  
-----------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- global level:


-- abstraction operator:

-- some remarks!!
--
-- in contrast to the original formulation in the paper, here we consider that
-- the states returned by abstract need not be flattened, so we perform some
-- additional calls to replace_flatten..
--
-- another difference with the original formulation in the paper: the states
-- recursively introduced by abstract may not follow the original rules
-- (e.g., that a ConsCall always has an empty stack) -> corrected
--
-- property: the state that results from a state (e,s) by removing e is always
-- flattened: if (e,(e2,x):stack) is flattened then (e2,stack) is also flattened..
-- check that this property is exploited everywhere.. also, some transformations
-- (like generalization, etc) may not destroy necessarily the flattened form
-- of states.. check it! (just to improve the efficiency of the slicing tool..)

abstract _ _ _ [] old = old
abstract pfuns funs n (e:es) old = 
  abstract pfuns funs n es (abs pfuns funs n e old)
  --abstract pfuns funs n es (abs pfuns funs n (trace ("\n____calling abs with:" ++ ppStates [e]) e) old)

-- abstraction for states:
--case expressions are not considered! the reason is that they must be reduced in 
--case_eval since they are always reducible..


--the second cases for var's lit's and conscall's were not in the paper (see above)
abs _ _ _ (Var _,[]) es = es
abs pfuns funs n (Var x,(s:ss)) es = abs pfuns funs n (replace_eval (Var x,(s:ss))) es

abs _ _ _ (Lit _,[]) es = es
abs pfuns funs n (Lit l,(s:ss)) es = abs pfuns funs n (replace_eval (Lit l,(s:ss))) es

abs pfuns funs n (Comb ConsCall _ args, []) es = 
  abstract pfuns funs n (replace_flatten pfuns funs n (addStacks args [])) es --different from the paper!!
abs pfuns funs n (Comb ConsCall c args,(s:ss)) es = abs pfuns funs n (replace_eval (Comb ConsCall c args,(s:ss))) es


abs pfuns funs n (Comb FuncCall f args, stack) es =
  if definedFunc2 (pfuns++funs) f
     then absFun pfuns funs n (Comb FuncCall f args, stack) es
     else abstract pfuns funs n ((removeExprFromState (Comb FuncCall f args, stack))
                                 : (replace_flatten pfuns funs n (addStacks args []))) es

abs pfuns funs n (Comb PartCall f args, stack) es = 
  addPartCall pfuns funs n (f,args) es stack

abs pfuns funs n (Apply e1 e2, s) es = 
  if (Apply e1 e2, s) `elem` es 
      then abstract pfuns funs n 
                    (replace_flatten pfuns funs n [(e1,[]),(e2,[]),
                              removeExprFromState (varBottom,s)]) es --varBottom is not important here..
      else (Apply e1 e2, s) : es

abs pfuns funs n (Or e1 e2,s) es =     -- this case is special..
  abstract pfuns funs n (replace_flatten pfuns funs n [(e1,s),(e2,s)]) es

abs pfuns funs n (GuardedExpr _ gc e, s) es = 
  abstract pfuns funs n (replace_flatten pfuns funs n [(gc,[]),(e,[]),removeExprFromState (varBottom,s)]) es
abs pfuns funs n (Constr _ e,s) es = abstract pfuns funs n (replace_flatten pfuns funs n [(e,[]),removeExprFromState (varBottom,s)]) es
abs pfuns funs n (Choice e,s) es = abstract pfuns funs n (replace_flatten pfuns funs n ([(e,[]),removeExprFromState (varBottom,s)])) es

---------------------------------------------
--auxiliary functions:

--for partial calls:
addPartCall pfuns funs n (f,e) es stack = 
  if definedFunc2 funs f
     then addPC pfuns funs (f,e) (getArityF funs [] f) es n stack
     else abstract pfuns funs n (replace_flatten pfuns funs n (addStacks e [])) es

addPC pfuns funs (f,e) i es n stack = 
  abstract pfuns funs n (replace_flatten pfuns funs n [(Comb FuncCall f (complete e i n),stack)]) es

--completes the arguments of a partcall with fresh variables
complete :: [Expr] -> Int -> Int -> [Expr]
complete [] n i = completeVars n i
complete (e:es) n i = e : (complete es (n-1) i)

completeVars :: Int -> Int -> [Expr]
completeVars n i = 
  if n==0 then [] else (Var ((maxExpVars*i)-n)) : (completeVars (n-1) i)


--for function calls:
absFun pfuns funs n (Comb FuncCall f e, stack) es = 
    if e_f == (varBottom,[])
       then flatState (EState pfuns funs (maxVarIndex (state2expr (Comb FuncCall f e, stack))) n) (Comb FuncCall f e, stack) : es
       else absFun_ pfuns funs n (Comb FuncCall f e, stack) e_f new_es
  where (new_es, e_f) = lookupExpr f es

--checks whether a function call already appear and returns
--the remaining calls and the existing call separately..
lookupExpr _ [] = ([],(varBottom,[]))
lookupExpr f ((exp,stack) : rest) =
  if isFun exp then lookupExprFun f ((exp,stack):rest)
               else lookupExpr_ f ((exp,stack):rest)

lookupExprFun f ((Comb FuncCall g es,s) : rest) =
  if f==g then (rest, (Comb FuncCall g es,s))
          else lookupExpr_ f ((Comb FuncCall g es,s) : rest)
          
lookupExpr_ f ((exp,s) : rest) = ((exp,s) : new_es, e_f)
  where (new_es, e_f) = lookupExpr f rest

--

absFun_ pfuns funs n (e,stack) (old_e,old_stack) es =   ---different from the paper!!
  if constructorInstance funs e [old_e]
     then abs pfuns funs n (removeExprFromState (e,stack)) ((old_e,old_stack):es)
     else absFun__ pfuns funs n (e,stack) (old_e,old_stack) es
--  if constructorInstance funs (state2expr e) [state2expr old_e]
--      then old_e : es
--      else absFun__ funs n e old_e es

removeExprFromState (_,[]) = (varBottom,[]) --will be removed in the next call to abs
removeExprFromState (_,(exp,_):stack) = (exp,stack)

absFun__ pfuns funs n (e,s1) (old_e,_) es =
    abstract pfuns funs n (replace_flatten pfuns funs n (addStacks ws [])) ((w,s1):es)
  where (w:ws) = msgT e old_e

state2expr (e,[]) = e
state2expr (e,(con,x):stack) = state2expr (substitute [x] [e] con, stack)

--
constructorInstance :: [FuncDecl] -> Expr -> [Expr] -> Bool
constructorInstance funs e es = 
  if sub==FSub then False else isConstructor funs sub
  where (sub,_) = search_instance e es

isConstructor :: [FuncDecl] -> Subst -> Bool
isConstructor funs (Sub _ expr) = andL (map (isCons funs) expr)
--this definition considers any non-defined symbol as constructor..
isCons _ (Var _) = True
isCons _ (Lit _) = True
isCons funs (Comb FuncCall f e) = if (definedFunc2 funs f)
  then False 
  else andL (map (isCons funs) e)
isCons funs (Comb ConsCall _ e) = andL (map (isCons funs) e)
isCons funs (Comb PartCall f e) = if (definedFunc2 funs f)
  then False 
  else andL (map (isCons funs) e)
isCons funs (Apply e1 e2) = andL (map (isCons funs) [e1,e2])
isCons funs (Constr _ e) = isCons funs e
isCons funs (Case _ e cases) = if (isCons funs e) 
  then isConsCase funs cases
  else False
isCons funs (Or e1 e2) = andL (map (isCons funs) [e1,e2])
isCons funs (GuardedExpr _ e1 e2) = andL (map (isCons funs) [e1,e2])
isCons funs (Choice e) = isCons funs e
--isCons _ (SQ _) = False  --isCons funs e

isConsCase _ [] = True
isConsCase funs (Branch _ e : les) =
  if (isCons funs e)
  then isConsCase funs les
  else False

--returns the matching substitution between an expression and a set of expressions
--(FSbub otherwise), as well as the matched term..
search_instance :: Expr -> [Expr] -> (Subst,Expr)
search_instance _ [] = (FSub, Var (-1))
search_instance e (e1:es) = 
  if sub==FSub then search_instance e es
               else (sub,e)
  where sub = instance e e1



--extracts the function calls in an expression:
extractFunCalls :: [Expr] -> [Expr]
extractFunCalls [] = []
extractFunCalls (Var _ :es) = extractFunCalls es
extractFunCalls (Lit _ :es) = extractFunCalls es
extractFunCalls (Comb ConsCall _ args :es) = extractFunCalls (args++es)
extractFunCalls (Comb PartCall _ args :es) = extractFunCalls (args++es)
extractFunCalls (Comb FuncCall f args :es) = 
  (Comb FuncCall f args ) : (extractFunCalls es)
extractFunCalls (Case _ e ces :es) = 
  (extractFunCalls [e]) ++ (extractFunCalls (map extractBranch ces)++es)
extractFunCalls (Apply e1 e2 : es) = 
  (extractFunCalls [e1,e2]) ++ (extractFunCalls es)
extractFunCalls (Or e1 e2 : es) = 
  (extractFunCalls [e1,e2]) ++ (extractFunCalls es)
extractFunCalls (Constr _ e : es) = 
  (extractFunCalls [e]) ++ (extractFunCalls es)
extractFunCalls (GuardedExpr _ e1 e2 : es) = 
  (extractFunCalls [e1,e2]) ++ (extractFunCalls es)
extractFunCalls (Choice e : es) = (extractFunCalls [e]) ++ (extractFunCalls es)

extractBranch (Branch _ e) = e



------------------------------------------------------------------------------
-- msg stuff

--msgT returns w U ran(theta) U ran(sigma)
--  where w = msg(e1,e2), e1 = theta(w), and e2 = sigma(w)
msgT :: Expr -> Expr -> [Expr]
msgT e f = (w:(subTerms sigma))++(subTerms theta)
  where (w,sigma,theta) = computeMsg e f
--kk1 = trace ("\ncomputeMSG: " ++ (ppExpr 0 w)) w

computeMsg :: Expr -> Expr -> (Expr,Subst,Subst)
computeMsg e1 e2 = (w, sub1, sub2)
  where (w,_) = msg e1 e2 (max (maxVarIndex e1) (maxVarIndex e2))
        sub1 = instance e1 w
        sub2 = instance e2 w

-- msg e1 e2 i = (w,i')
-- w is the msg of e1 and e2 and i' is the last number used in the 
-- generation of fresh variables during the computation of w 

msg :: Expr -> Expr -> Int -> (Expr,Int)
msg x y i =
  if x==y then (x, i)
     else if headExpr x == headExpr y
             then msgSameHead x y i
             else (Var (i+1), i+1)

headExpr :: Expr -> String
headExpr (Var _) = "Var"
headExpr (Lit _) = "Lit"
headExpr (Comb _ f _) = f
headExpr (Case Rigid _ _) = "Rigid"
headExpr (Case Flex _ _) = "Flex"
headExpr (Constr _ _) = "Constr"
headExpr (Choice _) = "Choice"
headExpr (Or _ _) = "Or"
headExpr (GuardedExpr _ _ _) = "GuardedExpr"
headExpr (Apply _ _) = "Apply"

msgSameHead :: Expr -> Expr -> Int -> (Expr, Int)

msgSameHead (Var x) (Var y) i =
  if x==y then (Var x, i)
          else (Var (i+1), i+1)
msgSameHead (Lit x) (Lit y) i =
  if x==y then (Lit x, i)
          else (Var (i+1), i+1)
msgSameHead (Comb c1 f1 e1) (Comb c2 f2 e2) i =
  if c1==c2 && f1==f2 then msgArgs c1 f1 e1 e2 [] i
                      else (Var (i+1), i+1)
msgSameHead (Case c1 e1 ces1) (Case c2 e2 ces2) i
  | c1==c2 && e1==e2 && ces1==ces2  = (Case c1 e1 ces1, i)
  | otherwise                       = (Var (i+1), i+1)
msgSameHead (Constr v1 e1) (Constr v2 e2) i =
  if v1==v2 && e1==e2 then (Constr v1 e1, i)
                      else (Var (i+1), i+1) --identical or generalize!
msgSameHead (Choice e1) (Choice e2) i =
  if e1==e2 then (Choice e1, i)
            else (Var (i+1), i+1) --identical or generalize!
msgSameHead (GuardedExpr v1 c1 e1) (GuardedExpr v2 c2 e2) i =
  if v1==v2 && c1==c2 && e1==e2 then (GuardedExpr v1 c1 e1, i) 
                                else (Var (i+1), i+1) --ident or gen!
msgSameHead (Or e1 e2) (Or f1 f2) i =
  if e1==f1 && e2==f2 then (Or e1 e2, i)
                      else (Var (i+1), i+1) --identical or generalize!
msgSameHead (Apply e1 e2) (Apply f1 f2) i =
  if e1==f1 && e2==f2 then (Apply e1 e2, i)
                      else (Var (i+1), i+1) --identical or generalize!

msgArgs :: CombType -> String -> [Expr] -> [Expr] -> [Expr] -> Int -> (Expr,Int)
msgArgs ctype f [] [] es i = (Comb ctype f es, i)
msgArgs ctype f (a:as) (b:bs) es i = msgArgs ctype f as bs (es++[ab]) j
  where (ab, j) = msg a b i


--given a term and a set, returns (Var -1) if there is no comparable (ie, with
--the same root symbol) term or the first comparable term and the remaining
--elements of the set:
comparable :: Expr -> [Expr] -> (Expr,[Expr])
comparable e es = comp e [] es

comp :: Expr -> [Expr] -> [Expr] -> (Expr,[Expr])
comp _ es [] = (Var (-1),es)
comp e es (h:t) =
  if headExpr e == headExpr h
     then (h, es++t)
     else comp e (es++[h]) t

-- end msg stuff
------------------------------------------------------------------------------
------------------------------------------------------------------------------

--extract vars and terms of a given (nonfailed) substitution
subVars :: Subst -> [Int]
subVars (Sub v _) = v
subVars FSub = trace "\n\n ERROR!!!!\n\n" [-2]
subTerms :: Subst -> [Expr]
subTerms (Sub _ e) = e
subTerms FSub = trace "\n\n ERROR!!!\n\n" [Var (-2)]


-- return the non-local variables of an expression

nonLocalVarsInExp (Var v) = [v]
nonLocalVarsInExp (Lit _) = []
nonLocalVarsInExp (Comb _ _ exps) = concat (map nonLocalVarsInExp exps)
nonLocalVarsInExp (Apply e1 e2) = nonLocalVarsInExp e1 ++ nonLocalVarsInExp e2
nonLocalVarsInExp (Constr vars e) = delArgs vars (nonLocalVarsInExp e)
nonLocalVarsInExp (Or e1 e2) = nonLocalVarsInExp e1 ++ nonLocalVarsInExp e2
nonLocalVarsInExp (Case _ e cases) = 
  nonLocalVarsInExp e ++ concat (map varsNonLocal cases)
nonLocalVarsInExp (GuardedExpr vars c e) = 
  delArgs vars (nonLocalVarsInExp c ++ nonLocalVarsInExp e)
nonLocalVarsInExp (Choice e) = nonLocalVarsInExp e
nonLocalVarsInExp (SQ e) = nonLocalVarsInExp e

-- return the non-local variables of the different scrutinees of a 
-- case expression
varsNonLocal (Branch (Pattern _ as) e) = delArgs as (nonLocalVarsInExp e) 
varsNonLocal (Branch (LPattern _) e)   = nonLocalVarsInExp e

-- transform variable indices into variable expressions and vice versa:
index2vars :: [VarIndex] -> [Expr]
index2vars = map (\i->Var i)

vars2index :: [Expr] -> [VarIndex]
vars2index = map (\(Var i)->i)

   
----------------------------------------------------------------------------
-- Local Control (unfolding, i.e., a non-standard metainterpreter for Curry)
----------------------------------------------------------------------------

-- local criteria to proceed unfolding a call at a given state:

--proceedUnfold pst _ = getDepth pst < 1

--selects the binding for some variable and returns its value:
selectBinding :: Int -> [(Int,Expr)] -> Expr
selectBinding _ [] = varBottom
selectBinding n ((x,b) : es) = if n==x then b else selectBinding n es

replaceBinding :: [(Int,Expr)] -> Int -> Expr -> [(Int,Expr)]
replaceBinding [] _ _ = []
replaceBinding ((y,b) : es) x exp = 
  if y==x then (y,exp) : replaceBinding es x exp
          else (y,b) : replaceBinding es x exp

removeBinding [] _ = []
removeBinding ((y,b) : es) x = 
  if y==x then es else (y,b) : removeBinding es x 

------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- case_eval implements the evaluation of case expressions

--case_eval :: Expr -> Expr

--this is necessary to terminate the evaluation of case expressions
--(and also when the initial expression is not a case expression..)
case_eval pst exp 
  --| trace ("\n\nCalling case_eval with: \n" ++ ppExpr 0 exp ++ "\n\n") False = varBottom
  | not(exp==exp)    = varBottom --to force the evaluation of exp...boh??
  | isCaseCons exp   = case_eval_ pst exp
  | isCaseLit exp    = case_eval_ pst exp
  | isCaseVar exp    = case_eval_ pst exp
  | isApply exp      = apply_eval pst exp
  --| isFun exp        = case_eval_fun pst exp
  | isBuiltin exp    = fun_eval pst exp
  | isOr exp         = or_eval pst exp
  | otherwise        = exp

--case_select:
case_eval_ pst (Case _ (Comb ConsCall f es) ces) = case_eval pst (matchBranch ces f es)

case_eval_ pst (Case _ (Lit l) ces) = case_eval pst (matchBranchLit ces l)

-- match a branch:
matchBranch :: [BranchExpr] -> String -> [Expr] -> Expr
matchBranch [] _ _ = (Comb FuncCall "failed" [])
matchBranch (Branch (Pattern p vars) e : ces) c es =
  if p==c then substitute vars es e else matchBranch ces c es

-- match a branch with Literals:
matchBranchLit :: [BranchExpr] -> Literal -> Expr
matchBranchLit [] _ = (Comb FuncCall "failed" [])
matchBranchLit (Branch (LPattern p) e : ces) c =
  if p==c then e else matchBranchLit ces c

--case_guess:

case_eval_ pst (Case _ (Var v) ces) = case_eval pst (propBindingNonDet v ces)

propBindingNonDet v (Branch (Pattern c vs) e : _) = 
   substitute [v] [(Comb ConsCall c (index2vars vs))] e
propBindingNonDet v (Branch (LPattern (Intc p)) e : _) = 
   substitute [v] [(Lit (Intc p))] e
propBindingNonDet v (Branch (LPattern (Floatc p)) e : _) = 
   substitute [v] [(Lit (Floatc p))] e
propBindingNonDet v (Branch (LPattern (Charc p)) e : _) = 
   substitute [v] [(Lit (Charc p))] e
propBindingNonDet v (_ : es) = propBindingNonDet v es  -- nondeterministic!!


--Higher-order:
--*** now after case_eval ***
--fun_eval pst (Apply (Comb PartCall f es) e2) =
--  if (getArity pst f)==(length es + 1)
--     then fun_eval pst (Comb FuncCall f (es++[e2]))
--     else Comb PartCall f (es++[e2])

--apply_eval pfuns funs exp 
--  | isApply exp   = apply_eval_ pfuns funs exp
--  | otherwise     = exp

apply_eval pst (Apply e1 e2)
  | isPartCall e1  = apply_eval_reduce pst (Apply e1 e2)
  | isApply e1     = apply_eval_try pst e1 e2
  | otherwise      = Apply e1 e2

apply_eval_try pst e1 e2 =
  if e1==new_e1 then Apply e1 e2
                else apply_eval pst (Apply new_e1 e2)
  where new_e1 = apply_eval pst e1

apply_eval_reduce pst (Apply (Comb PartCall f es) e2) =
  if (getArity pst f)==(length es + 1)
     then Comb FuncCall f (es++[e2])
     else Comb PartCall f (es++[e2])

or_eval pst (Or alt _) = case_eval pst alt
or_eval pst (Or _ alt) = case_eval pst alt

------------------------------------------------------------------------------
-- fun_eval implements the one-step unfolding of function calls 
--          (user-defined as well as those defined in the prelude)
--
--pending!! Constr, GuardedExpr, ... these constructions are currently
--          ignored here (and then removed by abstract)
--in principle, we don't want to have expressions rooted by GuardedExpr,
--Constr, etc, in the current set of expressions..


fun_eval :: EvalState -> Expr -> Expr

fun_eval pst exp 
  | isFun exp      = fun_eval_ pst exp
  | isApply exp    = apply_eval pst exp  --repeated (here and case_eval!!)
  | otherwise      = exp

fun_eval_ pst (Comb FuncCall f es) = 
  if definedFunc pst f
     then fun_eval_user pst f es
     else fun_eval_not_user pst f es

fun_eval_user pst f es = newexp
  where (_,newexp) = matchRHS pst f es

fun_eval_not_user pst f es = 
  if preludeFunc pst f
     then fun_eval_prelude pst f es
     else fun_eval_builtin f es

fun_eval_prelude pst f es = newexp
  where (_,newexp) = matchRHS_prelude pst f es

fun_eval_builtin f [e1,e2] =
  if isLit e1 && isLit e2 then fun_eval_builtin_ f e1 e2
                          else fun_eval_builtin__ f e1 e2

--builtin's:

fun_eval_builtin_ f (Lit (Intc x)) (Lit (Intc y))
  | f=="+"    = Lit (Intc (x+y))
  | f=="-"    = Lit (Intc (x-y))
  | f=="*"    = Lit (Intc (x*y))
  | f=="div"  = Lit (Intc (x `div` y))
  | f=="mod"  = Lit (Intc (x `mod` y))
  | f=="<"    = if x < y then Comb ConsCall "True" [] else Comb ConsCall "False" []
  | f==">"    = if x > y then Comb ConsCall "True" [] else Comb ConsCall "False" []
  | f=="<="   = if x <= y then Comb ConsCall "True" [] else Comb ConsCall "False" []
  | f==">="   = if x >= y then Comb ConsCall "True" [] else Comb ConsCall "False" []
  | otherwise = varBottom


--pending!! (==, =:=, &, etc)
fun_eval_builtin__ f e1 e2 
  | f=="=="    = fun_eval_builtin_eq e1 e2
  | otherwise  = Comb FuncCall f [e1,e2]
  
fun_eval_builtin_eq e1 e2 
  | isLit e1 && isLit e2           = if e1==e2 then Comb ConsCall "True" [] else Comb ConsCall "False" []
  | otherwise                      = Comb FuncCall "==" [e1,e2]


---------------------------------------------------------------------------
-- CALL UNFOLDING:

-- match a right-hand side of a given function:
matchRHS :: EvalState -> String -> [Expr] -> (EvalState,Expr)
matchRHS pst f es = getMatchedRHS pst (getFuncDecls pst) f es

-- match a right-hand side of a given function (defined in the prelude):
matchRHS_prelude :: EvalState -> String -> [Expr] -> (EvalState,Expr)
matchRHS_prelude pst f es = getMatchedRHS pst (getPreludeDecls pst) f es

getMatchedRHS :: EvalState -> [FuncDecl] -> String -> [Expr] -> (EvalState,Expr)
getMatchedRHS pst (Func fname _ _ funrule : fds) name es =
   if fname==name then getMatchedRHS_aux pst funrule es
                  else getMatchedRHS pst fds name es

getMatchedRHS_aux :: EvalState -> Rule -> [Expr] -> (EvalState,Expr)
getMatchedRHS_aux pst (Rule vars rhs) es =
  let curr_ri = getRenamingIndex pst
      n = getIter pst
      maxindex = max (maxList vars) (maxVarIndex rhs)
  in
   (updateRenamingIndex pst (curr_ri+maxindex+1),
    substituteAll vars es (curr_ri+maxExpVars*n) rhs)
   -- here we assume that no more than maxExpVars variables occur in each
   -- unfolded expression to produce a safe renaming.

-- (substitute vars exps expr) = expr[vars/exps]
-- i.e., replace all occurrences of vars by corresponding exps in the
-- expression expr
substitute :: [Int] -> [Expr] -> Expr -> Expr
substitute vars exps expr = substituteAll vars exps 0 expr

-- (substituteAll vars exps base expr):
-- substitute all occurrences of variables by corresponding expressions:
-- * substitute all occurrences of var_i by exp_i in expr
--   (if vars=[var_1,...,var_n] and exps=[exp_1,...,exp_n])
-- * substitute all other variables (Var j) by (Var (base+j))
--
-- here we assume that the new variables in guards and case patterns
-- do not occur in the list "vars" of replaced variables!

substituteAll :: [Int] -> [Expr] -> Int -> Expr -> Expr
substituteAll vars exps b (Var i) = replaceVar vars exps i
  where replaceVar [] [] var = Var (b+var)
        replaceVar (v:vs) (e:es) var = if v==var then e
                                                 else replaceVar vs es var
substituteAll _  _  _ (Lit l) = Lit l
substituteAll vs es b (Comb combtype c exps) =
                 Comb combtype c (map (substituteAll vs es b) exps)
substituteAll vs es b (Apply e1 e2) =
                 Apply (substituteAll vs es b e1) (substituteAll vs es b e2)
substituteAll vs es b (Constr vars e) =
                 Constr (map (+b) vars) (substituteAll vs es b e)
substituteAll vs es b (Or e1 e2) =
                 Or (substituteAll vs es b e1) (substituteAll vs es b e2)
substituteAll vs es b (Case ctype e cases) =
   Case ctype (substituteAll vs es b e) (map (substituteAllCase vs es b) cases)
substituteAll vs es b (GuardedExpr vars c e) =
   GuardedExpr (map (+b) vars) (substituteAll vs es b c)
                               (substituteAll vs es b e)
substituteAll vs es b (Choice e) = Choice (substituteAll vs es b e)
substituteAll vs es b (SQ e) = SQ (substituteAll vs es b e)

substituteAllCase :: [Int] -> [Expr] -> Int -> BranchExpr -> BranchExpr
substituteAllCase vs es b (Branch (Pattern l pvs) e) =
                 Branch (Pattern l (map (+b) pvs)) (substituteAll vs es b e)
substituteAllCase vs es b (Branch (LPattern l) e) =
                 Branch (LPattern l) (substituteAll vs es b e)


---------------------------------------------------------------------------
--replace rules (only for values):

replace_eval (Var y, (e,x):s) = (substitute [x] [Var y] e, s)
replace_eval (Comb ConsCall f es, (e,x):s) = (substitute [x] [Comb ConsCall f es] e, s)
replace_eval (Comb PartCall f es, (e,x):s) = (substitute [x] [Comb PartCall f es] e, s)
replace_eval (Lit l, (e,x):s) = (substitute [x] [Lit l] e, s)

--otherwise (nondeterministic!)
replace_eval (e,s) = (e,s)


---------------------------------------------------------------------------
---------------------------------------------------------------------------
--extractSlice constructs the slice

extractSlice _ [] _ = []
extractSlice (Prog x1 x2 x3 funs x4 x5) (e:es) all = 
    (lhs, red_rhs) : extractSlice (Prog x1 x2 x3 funs x4 x5) es all
  where (lhs, rhs, subs) = matchRule e funs
        red_rhs = reduce funs all subs rhs

matchRule e (Func fname _ _ funrule : fds) =
  if (topFun e)==fname
     then matchR e fname funrule
     else matchRule e fds

matchR (Comb _ _ es) fname (Rule vars rhs) =
  (Comb FuncCall fname (index2vars vars), rhs, Sub vars es)

topFun (Comb _ name _) = name


--reduce: reduces an expression without unfolding 
--(i.e., it computes the rhs of the slice)

reduce _ _ _ (Var x) = Var x
reduce _ _ _ (Lit l) = Lit l
reduce funs all subs (Comb ConsCall c es) = 
  Comb ConsCall c (map (reduce funs all subs) es)
reduce funs all (Sub vars exps) (Case ctype (Var x) ces) =
  reduce_case funs all (Sub vars exps) ctype (Var x) (substitute vars exps (Var x)) ces

--
reduce_case funs all subs ctype var (Comb ConsCall c es) ces =
  Case ctype var (reduce_matchBranch ces c es subs all funs)

reduce_matchBranch [] _ _ _ _ _ = [Branch (Pattern "Error" []) varBottom]
reduce_matchBranch (Branch (Pattern p vars) e : ces) c es subs all funs =
  if p==c then [Branch (Pattern p vars) (reduce funs all newsub e)]
          else reduce_matchBranch ces c es subs all funs
  where newsub = composeSubs subs (Sub vars es)

reduce_case funs all subs ctype var (Lit l) ces =
  Case ctype var (reduce_matchBranchLit ces l subs all funs)

reduce_matchBranchLit [] _ _ _ _ = [Branch (Pattern "Error" []) varBottom]
reduce_matchBranchLit (Branch (LPattern p) e : ces) c subs all funs =
  if p==c then [Branch (LPattern p) (reduce funs all subs e)]
          else reduce_matchBranchLit ces c subs all funs

reduce_case funs all subs ctype var (Var _) ces =
  Case ctype var (map (reduce_branch funs all subs) ces)

reduce_branch funs all subs (Branch p e) = Branch p (reduce funs all subs e)

reduce funs all subs (Comb FuncCall f es) =
  if definedFunc2 funs f 
     then reduce_fun funs all subs (Comb FuncCall f es)
     else Comb FuncCall f (map (reduce funs all subs) es)

reduce_fun funs all (Sub vars exps) (Comb FuncCall f es) =
    if sub==FSub
       then Comb ConsCall "T" []
       else Comb FuncCall f (map (reduce funs all (Sub vars exps)) es)
  where (sub,_) = search_fun f all
--  where (sub,_) = search_instance (substitute vars exps (Comb FuncCall f es)) all

search_fun _ [] = (FSub, Var (-1))
search_fun f (Comb _ g _ : es) = 
  if f==g then (Sub [] [], Var (-1))
          else search_fun f es


reduce funs all subs (Comb PartCall f es) =
  Comb PartCall f (map (reduce funs all subs) es)

reduce funs all subs (Apply e1 e2) = 
  Apply (reduce funs all subs e1) (reduce funs all subs e2)
reduce funs all subs (Or e1 e2) =
  Or (reduce funs all subs e1) (reduce funs all subs e2)
reduce funs all subs (Constr vars e) =
  Constr vars (reduce funs all subs e)
reduce funs all subs (GuardedExpr vars gc e) =
  GuardedExpr vars (reduce funs all subs gc) (reduce funs all subs e)
reduce funs all subs (Choice e) = Choice (reduce funs all subs e)


------------------------------------------------------------------------------
-- handling variables in expressions:


maxVarIndexHeap [] = -1
maxVarIndexHeap ((x,e):es) = 
  max (max (maxVarIndex (Var x)) (maxVarIndex e)) (maxVarIndexHeap es)

-- get the maximum index of all variables in an expression:
-- (or -1 if there is no variable)
maxVarIndex :: Expr -> Int
maxVarIndex (Var v)         = v
maxVarIndex (Lit _)         = -1
maxVarIndex (Comb _ _ exps) = maxList (map maxVarIndex exps)
maxVarIndex (Apply e1 e2)   = max (maxVarIndex e1) (maxVarIndex e2)
maxVarIndex (Constr vars e) = max (maxList vars) (maxVarIndex e)
maxVarIndex (Or e1 e2)      = max (maxVarIndex e1) (maxVarIndex e2)
maxVarIndex (Case _ ce cs)   = max (maxVarIndex ce) (maxList (map maxCase cs))
  where maxCase (Branch (Pattern _ pvs) e) = max (maxList pvs) (maxVarIndex e)
        maxCase (Branch (LPattern _)   e) = maxVarIndex e
maxVarIndex (GuardedExpr vars c e) =
                   max (max (maxList vars) (maxVarIndex c)) (maxVarIndex e)
maxVarIndex (Choice e)      = maxVarIndex e
maxVarIndex (SQ e)      = maxVarIndex e

-- maximum of two numbers:
max i j = if i<=j then j else i

-- maximum of a list of naturals (-1 if list is empty):
maxList xs = foldr max (-1) xs


--get the list of LOCAL variables in an expression:
localVarsInExp :: Expr -> [Int]
localVarsInExp e = diffList all_vars nonlvars
  where all_vars = varsInExp e
        nonlvars = nonLocalVarsInExp e

diffList :: [Int] -> [Int] -> [Int]
diffList [] [] = []
diffList (v:vs) vars =
  if v `elem` vars then diffList vs vars
                   else v : (diffList vs vars)


-- get list of all variables in an expression:
varsInExp :: Expr -> [Int]
varsInExp (Var v) = [v]
varsInExp (Lit _) = []
varsInExp (Comb _ _ exps) = concat (map varsInExp exps)
varsInExp (Apply e1 e2) = varsInExp e1 ++ varsInExp e2
varsInExp (Constr vars e) = vars ++ varsInExp e
varsInExp (Or e1 e2) = varsInExp e1 ++ varsInExp e2
varsInExp (Case _ e cases) = varsInExp e ++ concat (map varsInCase cases)
varsInExp (GuardedExpr vars c e) = vars ++ varsInExp c ++ varsInExp e
varsInExp (Choice e) = varsInExp e

varsInExp (SQ e) = varsInExp e

varsInCase (Branch (Pattern _ as) e) = as ++ varsInExp e
varsInCase (Branch (LPattern _)   e) = varsInExp e


-- delete argument variables in a list of variables
delArgs [] vs = vs
delArgs (x:xs) vs = delArgs xs (deleteAll x vs)

------------------------------------------------------------------------------
-- END local control 
------------------------------------------------------------------------------

------------------------------------------------------------------------------
-- Utilities 
------------------------------------------------------------------------------

--isSomething:

isPartCall :: Expr -> Bool
isPartCall x = case x of 
                    (Comb PartCall _ _) -> True
                    _ -> False

isApply :: Expr -> Bool
isApply x = case x of 
                (Apply _ _) -> True
                _ -> False


isConsCall :: Expr -> Bool
isConsCall x = case x of
                   (Comb ConsCall _ _) -> True
                   (Comb _ "success" _) -> True
                   (Comb _ "failed" _) -> True
                   _ -> False

isCaseVar :: Expr -> Bool
isCaseVar e = case e of
                   (Case _ (Var _) _) -> True
                   _ -> False

isCaseCons :: Expr -> Bool
isCaseCons e = case e of
                   (Case _ (Comb ConsCall _ _) _) -> True
                   _ -> False

isCaseLit :: Expr -> Bool
isCaseLit e = case e of 
                  (Case _ (Lit _) _) -> True
                  _ -> False

isCaseFun :: Expr -> Bool
isCaseFun e = case e of
                   (Case _ (Comb FuncCall _ _) _) -> True
                   _ -> False

isCaseApply :: Expr -> Bool
isCaseApply e = case e of
                    (Case _ (Apply _ _) _) -> True
                    _ -> False

-- is f a defined function?
definedFunc :: EvalState -> String -> Bool
definedFunc pst f = any (hasName f) (getFuncDecls pst)
definedFunc2 :: [FuncDecl] -> String -> Bool
definedFunc2 funs f = (any (hasName f) funs) --|| preludeFunc f

hasName :: String -> FuncDecl -> Bool
hasName name (Func fname _ _ _) = fname==name

-- is f a function of the prelude?
preludeFunc :: EvalState -> String -> Bool
preludeFunc pst f = any (hasName f) (getPreludeDecls pst)

--check whether an expression is a variable:
isVar :: Expr -> Bool
isVar x = case x of
              (Var _) -> True
              _ -> False

--check whether an expression is a Function Call:
isFun :: Expr -> Bool
isFun x = case x of
              (Comb FuncCall _ _) -> True
              _ -> False

--check whether an expression is a literal:
isLit :: Expr -> Bool
isLit x = case x of
               (Lit _) -> True
               _ -> False

--check whether an expression is rooted by Case:
isCase :: Expr -> Bool
isCase x = case x of
               (Case _ _ _) -> True
               _ -> False

--check whether an expression is rooted by Or:
isOr :: Expr -> Bool
isOr x = case x of
             (Or _ _) -> True
             _ -> False

isConstr :: Expr -> Bool
isConstr e = case e of
                 (Var _) -> True
                 (Lit _) -> True
                 (Comb ConsCall _ es) -> andL (map isConstr es)
                 _ -> False

-- compute arity of a function symbol:
getArity :: EvalState -> String -> Int
getArity pst f = getArityF (getFuncDecls pst) (getPreludeDecls pst) f

getArityF :: [FuncDecl] -> [FuncDecl] -> String -> Int
getArityF funs pfuns f
 | f=="+"    = 2
 | f=="-"    = 2
 | f=="*"    = 2
 | f=="div"  = 2
 | f=="mod"  = 2
 | f=="<"    = 2
 | f==">"    = 2
 | f=="<="   = 2
 | f==">="   = 2
 | f=="&"    = 2
 | f=="=:="  = 2
 | f=="=="   = 2
 | otherwise = findArity funs
 where
    findArity [] = findArityPrelude pfuns
    findArity (Func fn fa _ _ : fds) =
        if fn==f then fa else findArity fds
    findArityPrelude [] = error ("INTERNAL ERROR: cannot find arity of "++f)
    findArityPrelude (Func fn fa _ _ : fds) =
        if fn==f then fa else findArityPrelude fds



-- delete all occurrences of an element in a list:
deleteAll          :: a -> [a] -> [a]
deleteAll _ []     = []
deleteAll x (y:ys) = if x==y then deleteAll x ys else y : deleteAll x ys


-- check if two sets of terms are equal modulo renaming:
equalMR :: [Expr] -> [Expr] -> Bool
equalMR s1 s2 =  (existsRenSet s1 s2) && (existsRenSet s2 s1) 
 
--check if all the elements of the first set appear modulo renaming in the second set:
existsRenSet :: [Expr] -> [Expr] -> Bool
existsRenSet [] _ = True
existsRenSet (e:s) es = if (existsRen es e) then existsRenSet s es else False
  
--check if a term (second argument) appears (MR) in a list (first argument)
existsRen :: [Expr] -> Expr -> Bool

existsRen [] _ = False
existsRen (e1:es) e = 
  if  not (instance e1 e == FSub) &&  
      not (instance e e1 == FSub)
                      then True
                      else existsRen es e

-- encapsulated search..
-- (defined in the prelude as findFirst.. replace!)
first g = head (findall (\x -> g =:= x))

-- check whether one expression is an instance of another and compute
-- the corresponding substitution:
instance :: Expr -> Expr -> Subst
instance a b =
  if a==b then Sub [] []
          else if isVar b 
                  then instanceVar a b
                  else if headExpr a == headExpr b
                          then instanceS a b
                          else FSub

instanceVar :: Expr -> Expr -> Subst
instanceVar a (Var x) = Sub [x] [a]

--no need to treat vbles (otherwise we don't call to instanceS..)
instanceS :: Expr -> Expr -> Subst
instanceS (Lit x) (Lit y) = 
  if x==y then Sub [] [] else FSub
instanceS (Comb c1 f1 e1) (Comb c2 f2 e2) =
  if c1==c2 && f1==f2 then instanceL e1 e2 (Sub [] [])
                      else FSub
instanceS (Case c1 e1 ces1) (Case c2 e2 ces2) =
  if c1==c2 then instanceCase ces1 ces2 (instance e1 e2)
            else FSub
instanceS (Constr _ e1) (Constr _ e2) = instance e1 e2
instanceS (Choice e1) (Choice e2) = instance e1 e2
instanceS (GuardedExpr _ c1 e1) (GuardedExpr _ c2 e2) =
  instanceL [c1,e1] [c2,e2] (Sub [] [])
instanceS (Or e1 e2) (Or f1 f2) = instanceL [e1,e2] [f1,f2] (Sub [] [])
instanceS (Apply e1 e2) (Apply f1 f2) = instanceL [e1,e2] [f1,f2] (Sub [] [])

instanceL :: [Expr] -> [Expr] -> Subst -> Subst
instanceL [] [] s = s
instanceL (e1:es1) (e2:es2) subs 
  | newsub == FSub    =  FSub
  | clash newsub subs =  FSub
  | otherwise         =  instanceL es1 es2 (joinSubs subs newsub)
  where newsub = instance e1 e2
--  if newsub==FSub
--     then FSub
--     else instanceL es1 (map (substitute (subVars newsub) (subTerms newsub)) es2)
--                    (composeSubs subs newsub)
--  where newsub = instance e1 e2

clash (Sub vs es) sub2 = clash_ vs es sub2

clash_ [] [] _ = False
clash_ (v:vs) (e:es) (Sub vs2 es2) =
  if ve==varBottom then clash_ vs es (Sub vs2 es2) 
                   else not (e==ve)
  where ve = findVal v vs2 es2
  
findVal _ [] [] = varBottom
findVal w (v:vs) (e:es) = if w==v then e else findVal w vs es

--

instanceCase :: [BranchExpr] -> [BranchExpr] -> Subst -> Subst
instanceCase _ _ FSub = FSub
instanceCase ces1 ces2 (Sub sx vx) =
  instanceCaseL ces1 ces2 (Sub sx vx)
  --(map (substituteAllCase sx vx 0) ces2) (Sub sx vx)

instanceCaseL :: [BranchExpr] -> [BranchExpr] -> Subst -> Subst
instanceCaseL [] [] s = s
instanceCaseL (e1:es1) (e2:es2) subs 
  | newsub == FSub    =  FSub
  | clash newsub subs =  FSub
  | otherwise         =  instanceCaseL es1 es2 (joinSubs subs newsub)
  where newsub = instanceCaseBranch e1 e2
--  if newsub==FSub
--     then FSub
--     else instanceCaseL es1 
--          (map (substituteAllCase (subVars newsub) (subTerms newsub) 0) es2)
--          (composeSubs subs newsub)
--  where newsub = instanceCaseBranch e1 e2

instanceCaseBranch :: BranchExpr -> BranchExpr -> Subst
instanceCaseBranch (Branch (Pattern c1 _) e1) (Branch (Pattern c2 _) e2) = 
  if c1 == c2 then instance e1 e2
              else FSub
instanceCaseBranch (Branch (LPattern c1) e1) (Branch (LPattern c2) e2) = 
  if c1 == c2 then instance e1 e2
              else FSub


-- composition of substitutions:

composeSubs :: Subst -> Subst -> Subst
composeSubs (Sub vars1 exps1) (Sub vars2 exps2) = 
  (Sub (vars1 ++ vars2) (newExps ++ exps2))
  where newExps = (map (substitute vars2 exps2) exps1)
composeSubs FSub (Sub _ _) = FSub
composeSubs (Sub _ _) FSub = FSub
composeSubs FSub FSub = FSub

joinSubs (Sub v1 e1) (Sub v2 e2) = Sub (v1++v2) (e1++e2)

--conjunction of a list of boolean terms

andL:: [Bool] -> Bool
andL es = foldr (&&) True es

--disjunction of a list of boolean terms

orL:: [Bool] -> Bool
orL es = foldr (||) False es

--remove duplicated expressions (modulo variable renaming) 
removeRenamings :: [Expr] -> [Expr]
removeRenamings [] = []
removeRenamings (e:es) = if (existsRen es e) then removeRenamings es
                                            else e:(removeRenamings es)

--remove variables and literals from a list:
removeVarsLits :: [Expr] -> [Expr]
removeVarsLits [] = []
removeVarsLits (e:es) = 
  if (isVar e) || (isLit e) || (isConstr e)  then removeVarsLits es
                                             else e : (removeVarsLits es)
legalBuiltin c = case c of
  "=="           -> True
  "&"            -> True
  "&>"           -> True
  "=:="          -> True
  "*"            -> True
  "+"            -> True
  "-"            -> True
  "div"          -> True
  "mod"          -> True
  "<"            -> True
  ">"            -> True
  "<="           -> True
  ">="           -> True
  "findall"      -> True
  "failed"       -> True
  "readFlatCurry"-> True
  "trace"        -> True
  _              -> False

isBuiltin e = case e of
                  (Comb FuncCall c _) -> legalBuiltin c
                  _ -> False

------------------------------------------------------------------------------
-- END utilities 
------------------------------------------------------------------------------

------------------------------------------------------------------------------
-- Pretty print 
------------------------------------------------------------------------------

delMod [] = []
delMod (s:ss) = if s=='.' then ss else delMod ss

delMod2 ss = if '.' `elem` ss then delMod ss else ss

-- pretty printer for expressions (first argument: indentation):
ppExpr _ (Var n) = ppVar n
ppExpr _ (Lit l) = ppLit l
ppExpr b (Comb ConsCall cf es) =
  "(" ++ (delMod2 cf) ++ " " ++ ppList (ppExpr b) es ++ ")"
ppExpr b (Comb PartCall cf es) =
  "(" ++ (delMod cf) ++ " " ++ ppList (ppExpr b) es ++ ")"
ppExpr b (Comb FuncCall cf es) =
  "(" ++ (delMod cf) ++ " " ++ ppList (ppExpr b) es ++ ")"
ppExpr b (Apply e1 e2) =
  "(" ++ ppExpr b e1 ++ " " ++ ppExpr b e2 ++ ")"
ppExpr b (Constr xs e) =
  "(Constr " ++ ppFormatList ppVar xs ++ (ppExpr b) e ++ ")"
ppExpr b (Or e1 e2) =
  "(Or " ++ ppExpr b e1 ++ " " ++ ppExpr b e2 ++ ")"
ppExpr b (Case Rigid e cs) =
  "(Case " ++ ppExpr b e ++ " of\n " ++ ppList (ppCase (b+2)) cs
  ++ blanks b ++ ")"
ppExpr b (Case Flex e cs) =
  "(FCase " ++ ppExpr b e ++ " of\n " ++ ppList (ppCase (b+2)) cs
  ++ blanks b ++ ")"
ppExpr b (GuardedExpr xs e1 e2) =
  "(GuardedExpr " ++ ppFormatList ppVar xs
                  ++ ppExpr b e1 ++ " " ++ ppExpr b e2 ++ ")"
ppExpr b (Choice e) = "(Choice " ++ ppExpr b e ++ ")"

ppExpr b (SQ e) = "(SQ " ++ ppExpr b e ++ ")"

ppVar i = "x" ++ show i

ppLit (Intc   i) = show i
ppLit (Floatc f) = show f
ppLit (Charc  c) = "'" ++ show (ord c) ++ "'"

ppCase b (Branch (Pattern l xs) e) = blanks b ++ "(" ++ l ++ " "
                                     ++ ppList ppVar xs
                                     ++ " -> " ++ ppExpr b e ++ ")\n"
ppCase b (Branch (LPattern l) e) = blanks b ++ "(" ++ ppLit l ++ " "
                                   ++ " -> " ++ ppExpr b e ++ ")\n"

blanks b = take b (repeat ' ')

-- format a finite list of elements (with a format function for list elems):
ppFormatList :: (a->String) -> [a] -> String
ppFormatList format elems = " [" ++ ppElems elems ++ "] "
 where
   ppElems [] = ""
   ppElems [x] = format x
   ppElems (x1:x2:xs) = format x1 ++ "," ++ ppElems (x2:xs)

ppList format elems = ppListElems elems
  where
    ppListElems [] = ""
    ppListElems [x] = format x
    ppListElems (x1:x2:xs) = format x1 ++ " " ++ ppListElems (x2:xs)


      
ppResultants :: [(Expr,Expr)] -> IO ()
ppResultants [] = done
ppResultants ((lhs,rhs):es) = 
  do putStrLn (ppExpr 0 lhs ++ " = " ++ ppExpr 0 rhs)
     ppResultants es

ppRen:: [(Expr,Expr)] -> IO ()
ppRen [] = putStr "\n"
ppRen ((e1,e2):es) = 
  do putStrLn (ppExpr 0 e1 ++ "->" ++ ppExpr 0 e2)
     ppRen es


pprenLHS :: Expr -> [(Expr,Expr)] -> Expr
pprenLHS e ((s,ns):rest) = 
  if e == s then ns else pprenLHS e rest

-- pretty print for substitutions..
pprintSub FSub = putStrLn "FSub"
pprintSub (Sub vars exps) =
   putStrLn (concatMap show vars) >>
   putStrLn (concat (map (ppExpr 0) exps))


-- pprint: prints a program in FlatCurry form

pprint :: String -> IO ()
pprint p = do (Prog _ _ _ fds _ _) <- readFlatCurry p
              ppFuncDecls fds

ppFuncDecls :: [FuncDecl] -> IO ()
ppFuncDecls fds = foldr (>>) done (map ppFunRule fds)
  where
    ppFunRule (Func fname _ _ rule) = ppRule fname rule

-- pretty printer for rules:
ppRule :: String -> Rule -> IO ()
ppRule fname (Rule lhsVars rhs) =
   putStrLn (fname ++ " " ++ ppList ppVar lhsVars ++ " = " ++ ppExpr 0 rhs)
ppRule fname (External ename) =
   putStrLn (fname ++ " external \"" ++ ename ++ "\"")


------------------------------------------------------------------------------
-- END pretty print 
------------------------------------------------------------------------------


-- end of program
